
#include <sys/types.h>
#include <unistd.h>
#include <sys/mman.h>
#include <limits.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>

/* This seems wrong.  I think this is an ordinary POSIX feature, not rtems. We can't define this up front because the system types get very confused.  So, we set this only for sys/signal.h */
#define __rtems__
#define _POSIX_REALTIME_SIGNALS
#undef _SYS_SIGNAL_H
#include <sys/signal.h>
#undef __rtems__

#include "page_table.h"
#include "debug.h"
#include "list.h"

struct page_table {
	char *base;
	int pages;
	int *access;
	page_table_handler_t handler;
	void *handler_arg;
};

static int page_size=0;
static struct list *table_list=0;

static int convert_to_unix( int access );
static struct page_table * find_address( char *addr, int *page );
static void fault_handler( int num, siginfo_t *info, void *context );

struct page_table * page_table_create( char *base, int pages, int access, page_table_handler_t handler, void *handler_arg )
{
	int i;
	struct sigaction sa;
	struct page_table *pt;

	page_table_pagesize();

	if(!table_list) {
		table_list = list_create();
		if(!table_list) return 0;
	}

	pt = malloc(sizeof(struct page_table));
	if(!pt) return 0;

	pt->base = base;
	pt->pages = pages;
	pt->access = malloc(sizeof(int)*pages);
	pt->handler = handler;
	pt->handler_arg = handler_arg;

	for( i=0; i<pt->pages; i++ ) {
		pt->access[i] = access;
	}

	list_push_head(table_list,pt);

	sa.sa_sigaction = fault_handler;
	sigfillset( &sa.sa_mask );
	sa.sa_flags = SA_SIGINFO;

	sigaction( SIGSEGV, &sa, 0 );

	mprotect(base,page_size*pages,convert_to_unix(access));

	return pt;
}

void page_table_delete( struct page_table *pt )
{
	mprotect(pt->base,page_size*pt->pages,convert_to_unix(PAGE_TABLE_ACCESS_ALL));
	list_remove(table_list,pt);
	free(pt->access);
	free(pt);
}

int page_table_access_set( struct page_table *pt, int page, int npages, int access )
{
	if( page<0 || ((page+npages) > pt->pages) ) {
		abort();
		return 0;
	} else {
		int i;
		int result;
		for( i=0; i<npages; i++ ) {
			pt->access[page+i] = access;
		}
		result = mprotect( pt->base+page*page_size, page_size*npages, convert_to_unix(access) );
		if(result!=0) abort();
		return 1;
	}

}

int page_table_access_get( struct page_table *pt, int page )
{
	if( page<0 || (page>=pt->pages) ) {
		abort();
		return 0;
	} else {
		return pt->access[page];
	}
}

char *page_table_base( struct page_table *pt )
{
	return pt->base;
}

struct page_table * page_table_from_base( char *addr )
{
	struct list_node *i;

	if(!table_list) return 0;

	for(i=table_list->head;i;i=i->next) {
		struct page_table *t = i->data;
		if( addr==t->base ) return t;
	}

	return 0;
}


char *page_table_data( struct page_table *pt, int page )
{
	return pt->base + page*page_size;
}

int page_table_pages( struct page_table *pt )
{
	return pt->pages;
}

int page_table_pagesize()
{
	if(page_size==0) {
		#ifdef _SC_PAGE_SIZE
		page_size = sysconf(_SC_PAGE_SIZE);
		#else
		page_size = sysconf(_SC_PAGESIZE);
		#endif
	}

	return page_size;
}

static int convert_to_unix( int access )
{
	int result=0;

	if(access&PAGE_TABLE_ACCESS_READ) result |= PROT_READ;
	if(access&PAGE_TABLE_ACCESS_WRITE) result |= PROT_WRITE;
	if(access&PAGE_TABLE_ACCESS_EXEC) result |= PROT_EXEC;

	return result;
}

static struct page_table * find_address( char *addr, int *page )
{
	struct list_node *i;

	if(!table_list) return 0;

	for(i=table_list->head;i;i=i->next) {
		struct page_table *t = i->data;
		if( addr<t->base ) continue;
		*page = (addr-t->base ) / page_size;
		if( *page < t->pages ) return t;
	}

	return 0;
}

#ifdef linux

#include <asm/sigcontext.h>
#include <asm/ucontext.h>

char *get_fault_address( void *context )
{
	return (char*)(((struct ucontext *)context)->uc_mcontext.cr2);
}


#endif

static void fault_handler( int num, siginfo_t *info, void *context )
{
	struct page_table *t;
	void *addr;
	int page;

	#if defined(__CYGWIN__)
		addr = info->si_value.sival_ptr;
	#elif defined(linux)
		addr = get_fault_address(context);
	#else
		addr = info->si_addr;
	#endif

	t = find_address( addr, &page );

	if( !t || !t->handler ) {
		fprintf(stderr,"Segmentation fault at address %x\n",(unsigned)addr);
		abort();
	} else {
		t->handler(t,page,t->handler_arg);
	}
}

int page_table_iterate( page_table_iterator_t iter, void *arg )
{
	if(table_list) {
		return list_iterate( table_list, (list_op_t) iter, arg );
	} else {
		return 1;
	}
}
