/*
dttools is Copyright (C) 2002 Douglas Thain
This software is released under the GNU General Public License.
See the file COPYING for details.
*/


#include "sparse_file.h"

#include "full_io.h"
#include "safe_file.h"
#include "interval_set.h"
#include "create_dir.h"
#include "delete_dir.h"
#include "pvalue.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>

#define SFLF SPARSE_FILE_LENGTH_FORMAT

struct sparse_file {
	char dirname[PATH_MAX];
	struct interval_set *intervals;
	sparse_file_length_t size;
	int size_known;
	int fd;
};

struct sparse_file * sparse_file_open( const char *dirname )
{
	struct sparse_file *h;
	char path[PATH_MAX];
	FILE *file;

	h = malloc(sizeof(*h));
	if(!h) return 0;

	memset(h,0,sizeof(*h));

	h->intervals = interval_set_create(); 
	if(!h->intervals) goto failure;

	strcpy(h->dirname,dirname);

	sprintf(path,"%s/data",dirname);
	h->fd = open(path,O_RDWR,0777);
	if(h->fd<0) goto failure;

	sprintf(path,"%s/intervals",dirname);
	file = fopen(path,"r");
	if(file) {
		if(!interval_set_load(h->intervals,file)) {
			fclose(file);
			goto failure;
		}
		fclose(file);
	}

	if(sparse_file_get_attr(h,"size",SFLF,&h->size)) {
		h->size_known = 1;
	} else {
		h->size_known = 0;
	}

	return h;

	failure:
	sparse_file_close(h);
	return 0;
}

struct sparse_file * sparse_file_create( const char *dirname )
{
	struct sparse_file *h;
	char path[PATH_MAX];
	FILE *file;

	h = malloc(sizeof(*h));
	if(!h) return 0;

	memset(h,0,sizeof(*h));

	h->intervals = interval_set_create(); 
	if(!h->intervals) goto failure;

	strcpy(h->dirname,dirname);
	if(!create_dir(dirname)) goto failure;

	sprintf(path,"%s/data",dirname);
	h->fd = open(path,O_CREAT|O_TRUNC|O_RDWR,0777);
	if(h->fd<0) goto failure;

	sprintf(path,"%s/intervals",dirname);
	file = fopen(path,"w");
	if(!file) goto failure;
	if(!interval_set_save(h->intervals,file)) {
		fclose(file);
		goto failure;
	}
	fclose(file);

	h->size_known = 0;
	h->size = 0;

	return h;

	failure:
	sparse_file_close(h);
	sparse_file_delete(dirname);
	return 0;
}

void sparse_file_close( struct sparse_file *h )
{
	if(h) {
		if(h->fd>0) close(h->fd);
		interval_set_delete(h->intervals);
		free(h);
	}
}

void sparse_file_delete( const char *path )
{
	delete_dir(path);
}

char * sparse_file_name( struct sparse_file *s )
{
	return s->dirname;
}

int  sparse_file_set_size( struct sparse_file *h, sparse_file_length_t size )
{
	/* Physical truncation is an optimization. */
	/* If it fails, no big deal. */

	ftruncate(h->fd,size);

	/* However, the success of logical truncation */
	/* must be checked carefully. */

	if(sparse_file_set_attr(h,"size",SFLF,size)) {
		h->size_known = 1;
		h->size = size;
		return 1;
	} else {
		return 0;
	}
}

int  sparse_file_get_size( struct sparse_file *h, sparse_file_length_t *size )
{
	if(h->size_known) {
		*size = h->size;
		return 1;
	} else {
		return 0;
	}
}

void sparse_file_get_nbytes( struct sparse_file *h, sparse_file_length_t *nbytes )
{
	interval_set_nbytes(h->intervals,nbytes);
}

int  sparse_file_read( struct sparse_file *h, void *data, sparse_file_length_t length, sparse_file_length_t offset, sparse_file_length_t *actual )
{
	int result;

	if(h->size_known) {
		if(offset>=h->size) {
			*actual = 0;
			return 1;
		}
	}

	result = interval_set_contains(h->intervals,offset,length,actual);
	if(!result) return 0;

	result = full_pread(h->fd,data,*actual,offset);
	return result==*actual;
}

int  sparse_file_write( struct sparse_file *h, const void *data, sparse_file_length_t length, sparse_file_length_t offset )
{
	int result;

	result = full_pwrite(h->fd,data,length,offset);
	if(result!=length) return 0;

	return interval_set_add(h->intervals,offset,length);
}

int  sparse_file_commit( struct sparse_file *h )
{
	struct safe_file *file;
	char path[PATH_MAX];
	int result;

	result = fsync(h->fd);
	if(result!=0) return 0;

	sprintf(path,"%s/intervals",h->dirname);

	file = safe_file_begin(path);
	if(!file) return 0;

	if(!interval_set_save(h->intervals,safe_file_stream(file))) {
		safe_file_abort(file);
		return 0;
	}
 
	if(safe_file_commit(file)) {
		return 1;
	} else {
		return 0;
	}
}

int sparse_file_set_attr( struct sparse_file *h, const char *name, const char *fmt, ... )
{
	char path[PATH_MAX];

	va_list args;
	va_start(args,fmt);

	sprintf(path,"%s/%s",h->dirname,name);
	return pvalue_vset(path,fmt,args);

	va_end(args);
}

int sparse_file_get_attr( struct sparse_file *h, const char *name, const char *fmt, ... )
{
	char path[PATH_MAX];

	va_list args;
	va_start(args,fmt);

	sprintf(path,"%s/%s",h->dirname,name);
	return pvalue_vget(path,fmt,args);

	va_end(args);
}


